package org.fmi.ejb.impl;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import javax.ejb.EJBException;
import javax.ejb.Remote;
import javax.ejb.Stateful;
import javax.ejb.Stateless;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.Reference;
import javax.rmi.PortableRemoteObject;

import org.fmi.ejb.Access;
import org.fmi.ejb.Callback;
import org.fmi.ejb.EJB;
import org.fmi.ejb.EJBContainer;
import org.fmi.ejb.EJBObjectFactory;
import org.fmi.ejb.IIOPRemoteObjectFactory;
import org.fmi.ejb.deployer.Deployer;

public class EJBContainerImpl implements EJBContainer {
	
	private static final PrintStream PS = System.out;
	private Deployer deployer;
	private IIOPRemoteObjectFactory iiopAccessFactory; 
	
	private List<DeploymentInfo> deploymentInfo;
	private Map<String,EJB> ejbs;
//	private ORB orb;
	
	
	EJBContainerImpl(File deployDir) {
		deployer = new Deployer(deployDir,this);
		deploymentInfo = new ArrayList<DeploymentInfo>();
		ejbs = new HashMap<String,EJB>();
		instance = this;
		initIIOPAccess();
	}
	
	
	private void initIIOPAccess() {
		try {
			InitialContext ictx = new InitialContext();
			String host = System.getProperty("org.fmi.ejb.orbhost","localhost");
			String port = System.getProperty("org.fmi.ejb.orbport","900");
			Context ctx = (Context)ictx.lookup("corbaname:iiop:1.2@" + host + ":" + port + "#/");
			iiopAccessFactory = new EJBObjectFactory();
			PortableRemoteObject.exportObject(iiopAccessFactory);
			ctx.rebind("IIOPRemoteObjectFactory",iiopAccessFactory);
		} catch(Exception ex) {
			throw new EJBException(ex);
		}
	}
	
	public void start() {
		PS.print("Starting EJB Container...");
		Thread t = new Thread(deployer,"EJB Deployer");
		t.setDaemon(true);
		t.start();
		PS.println(" done.");
	}

	public void stop() {
		PS.print("Stopping EJB Container...");
		deployer.stop();
		PS.println(" done.");
	}

	public Properties getEJBProperties() {
		// TODO Auto-generated method stub
		return null;
	}

	public EJB getEJB(String name) {
		return ejbs.get(name);
	}

	public Collection<EJB> getAllEJB() {
		// TODO Auto-generated method stub
		return null;
	}

	public EJB getEJB(Reference ref) {
		// TODO Auto-generated method stub
		return null;
	}

	public void deploy(File jar) {
		PS.println("Start deploying '" + jar + "' ... ");
		DeploymentInfo info = new DeploymentInfo(jar);
		deploymentInfo.add(info);
		startEJBs(info);
		PS.println("End deploying '" + jar + "' ... ");
	}
	
	private void startEJBs(DeploymentInfo info) {
		Map<String,Class> stateless = info.getStatelessBeans();
		for (String beanName:stateless.keySet()) {
			Class clazz = stateless.get(beanName);
			assert clazz != null;
			deployBean(beanName,clazz,true,info);
		}
		Map<String,Class> stateful = info.getStatefulBeans();
		for (String beanName:stateful.keySet()) {
			Class clazz = stateful.get(beanName);
			assert clazz != null;
			deployBean(beanName,clazz,false,info);
		}
	}
	
	
	private void deployBean(String name,Class clazz,boolean isStateless,DeploymentInfo info) {
		DefaultEJB ejb = new DefaultEJB(name,clazz,info.getClassLoader(),!isStateless);
		ejb.setChainsManager(new DefaultChainsManager(ejb));
		ejb.setClientViewsManager(new DefaultClientViewsManager(ejb));
		if (isStateless) {
			ejb.setLifecycleManager(new StatelessLifecycleManager(ejb));
		} else {
			ejb.setLifecycleManager(new StatefulLifecycleManager(ejb));
		}
		ejbs.put(name,ejb);		
		
		// init remote client view factories and invocation chains
		Remote remoteAnnotation = (Remote)ejb.getBeanClass().getAnnotation(Remote.class);
		if ( remoteAnnotation != null ) {
			for (Class remoteInterface:remoteAnnotation.value()) {
				// create factory
				RemoteClientViewFactory factory = new RemoteClientViewFactory(Access.REMOTE,
						new RemoteEJBProxyInvocationHandlerFactory(),remoteInterface,ejb);
				ejb.getClientViewsManager().registerFactory(factory);
				
				DefaultInvocationChains chains = new DefaultInvocationChains(remoteInterface);
				
				//add all chains
				for (Method method:remoteInterface.getMethods()) {
					List<Integer> interceptors = new ArrayList<Integer>();
					// TODO 
					interceptors.add(ejb.getChainsManager().register(
							new Interceptor_StatelessInstanceGetter()));
					DefaultInvocationChain chain = new DefaultInvocationChain(interceptors,
							getBusinessMethod(method,ejb.getBeanClass()));
					chains.addChain(method,chain);
				}
				//add chains for creation due to dependency injection
				List<Integer> interceptors = new ArrayList<Integer>();
				if (isStateless) {
					interceptors.add(ejb.getChainsManager().register(
							new Interceptor_StatelessProxyInstanceCreation(remoteInterface.getName())));
				} else {
					// TODO 
				}
				DefaultInvocationChain chain = new DefaultInvocationChain(interceptors,null);	
				chains.addChain(null,chain);
				
				ejb.getChainsManager().register(chains);
			}
		}
		// init local client view factories and invocation chains
		
		// init lifecycle invocation chains
		DefaultCallbackInvocationChains callbackchains = new DefaultCallbackInvocationChains();
		if (isStateless) {
			// NEW 
			List<Integer> interceptors = new ArrayList<Integer>();
			interceptors.add(ejb.getChainsManager().register(new Interceptor_NewStatelessContext()));
			interceptors.add(ejb.getChainsManager().register(new Interceptor_ApplicationInstancesCreator()));
			DefaultInvocationChain chain = new DefaultInvocationChain(interceptors,null);
			callbackchains.addChain(Callback.NEW,chain);
			// DESTROY
			// DISCARD
		} else {
			
		}
		ejb.getChainsManager().register(callbackchains);
	}
	
	private Method getBusinessMethod(Method intf, Class<?> beanClass) {
		try {
			return beanClass.getDeclaredMethod(intf.getName(),(Class[])intf.getParameterTypes());
		} catch(Exception ex) {
			throw new EJBException();
		}
	}
	
	/*
	 * MAIN method
	 */
	public static void main(String[] args) throws Exception {
		EJBContainerImpl container = new EJBContainerImpl(
				new File(".","deploy"));
		final Object LOCK = new Object();
		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable(){
			public void run() {
				synchronized(LOCK) {
					LOCK.notifyAll();
				}
			}
		}));
		container.start();
		synchronized (LOCK) {
			LOCK.wait();
		}
		container.stop();
	}
	
	
	private static class DeploymentInfo {
		
		private EJBClassLoader loader;
		private File file;
		private Map<String,Class> stateless;
		private Map<String,Class> stateful;
		
		DeploymentInfo(File file) {
			this.file = file;
			try {
				this.loader = new EJBClassLoader(
						new URL[]{file.toURL()},DeploymentInfo.class.getClassLoader());
			} catch(MalformedURLException ex) {
				throw new EJBException(ex);
			}
			stateless = new HashMap<String,Class>();
			stateful = new HashMap<String,Class>();
			initBeans();
			List<String> names = new ArrayList<String>();
			names.addAll(stateless.keySet());
			names.addAll(stateful.keySet());
			loader.setBeanNames(names);
		}
		
		public Map<String,Class> getStatelessBeans() {
			return Collections.unmodifiableMap(stateless);
		}
		
		public Map<String,Class> getStatefulBeans() {
			return Collections.unmodifiableMap(stateful);
		}
		
		public EJBClassLoader getClassLoader() {
			return loader;
		}
		
		private void initBeans() {
			try {
				JarFile jar = new JarFile(file);
				Enumeration<JarEntry> entries = jar.entries();
				while (entries.hasMoreElements()) {
					JarEntry entry = entries.nextElement();
					if ( entry.isDirectory() ) {
						continue;
					}
					String name = entry.getName();
					if ( !name.endsWith(".class") ) {
						continue;
					}
					
					String className = name.substring(0,name.length() - 
							".class".length()).replace("/",".");
					Class<?> clazz = loader.loadClass(className);
					
					Stateless sl = null;
					if ( (sl = (Stateless)clazz.getAnnotation(Stateless.class)) != null ) { 
						String beanName = null;
						if ( sl.name() != null && !sl.name().equals("")) {
							beanName = sl.name();
						} else {
							beanName = className.substring(
									className.lastIndexOf(".") + 1,
									className.length());
						}
						stateless.put(beanName,clazz);
						PS.println(" found stateless bean '" + beanName + "' in class " +
								className);
						continue;
					}
					Stateful sf = null;
					if ( (sf = (Stateful)clazz.getAnnotation(Stateful.class)) != null ) { 
						String beanName = null;
						if ( sf.name() != null && !sf.name().equals("")) {
							beanName = sf.name();
						} else {
							beanName = className.substring(
									className.lastIndexOf(".") + 1,
									className.length());
						}
						stateful.put(beanName,clazz);
						PS.println(" found stateless bean '" + beanName + "' in class " +
								className);
						continue;
					}
				}
			} catch(NoClassDefFoundError er) {
				throw new EJBException(new RuntimeException(er));
			} catch(ClassNotFoundException ex) {
				throw new EJBException(ex);
			} catch(IOException ex) {
				throw new EJBException(ex);
			}
		}
		
	}
	
	static EJBContainer instance;

	public static EJBContainer getInstance() {
		return instance;
	}
}
